/* slld_hal_zynq7000.c - SLLD Hardware Abstraction Layer Code for Xilinx Zynq 7000 QSPI */

/**************************************************************************
 * 2003-2016, Cypress Semiconductor Corporation or a subsidiary of Cypress 
 * Semiconductor Corporation.  All rights reserved.
 * 
 * This software, including source code, documentation and related materials 
 * (Software), is owned by Cypress Semiconductor Corporation or one of its 
 * subsidiaries (Cypress) and is protected by and subject to worldwide patent 
 * protection (United States and foreign), United States copyright laws and 
 * international treaty provisions. Therefore, you may use this Software only 
 * as provided in the license agreement accompanying the software package from 
 * which you obtained this Software (EULA).
 * 
 * If no EULA applies, Cypress hereby grants you a personal, non-exclusive, 
 * non-transferable license to copy, modify, and compile the Software source code 
 * solely for use in connection with Cypress's integrated circuit products.  Any 
 * reproduction, modification, translation, compilation, or representation of 
 * this Software except as specified above is prohibited without the express 
 * written permission of Cypress.
 * 
 * Disclaimer: THIS SOFTWARE IS PROVIDED AS-IS, WITH NO WARRANTY OF ANY KIND, 
 * EXPRESS OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, NONINFRINGEMENT, IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. Cypress 
 * reserves the right to make changes to the Software without notice. Cypress does 
 * not assume any liability arising out of the application or use of the Software 
 * or any product or circuit described in the Software. Cypress does not authorize 
 * its products for use in any products where a malfunction or failure of the Cypress 
 * product may reasonably be expected to result in significant property damage, injury 
 * or death (High Risk Product). By including Cypress's product in a High Risk Product, 
 * the manufacturer of such system or application assumes all risk of such use and in 
 * doing so agrees to indemnify Cypress against all liability.
 */

#include <stdio.h>
#include "slld_targetspecific.h"
#include "slld.h"
#include "slld_hal.h"

void DelayMicroseconds(int microseconds);

// Enable this define to turn on MIOs for QSPI1, otherwise we only use QSPI0
//#define DUAL_QSPI

// ***************************************************************************
//  FLASH_INIT - HAL initialization function
//
//  input : none
//
//  return value : status of the initialization - FAIL or SUCCESS
// ***************************************************************************
SLLD_STATUS FLASH_INIT
(
void
)
{
    DWORD       dwRegisterData;
    int         i;

    // Check the QSPI controller clock
    dwRegisterData = SLCR_LQSPI_CLK_CTRL_REG;
    if (dwRegisterData != QSPI_CLK_ARM_PLL_DIV7_VALUE)
    {
        printf("\n Updating QSPI clock control register from %08lxh to  %08lxh",
                                        dwRegisterData, QSPI_CLK_ARM_PLL_DIV7_VALUE);
        SLCR_LQSPI_CLK_CTRL_REG = QSPI_CLK_ARM_PLL_DIV7_VALUE;
    }

#ifdef DUAL_QSPI
    for (i = MIO_START_PIN_NUMBER_DUAL_QSPI; i <= MIO_END_PIN_NUMBER_DUAL_QSPI; i++)
#else
    for (i = MIO_START_PIN_NUMBER_SINGLE_QSPI; i <= MIO_END_PIN_NUMBER_SINGLE_QSPI; i++)
#endif
    {
        dwRegisterData = *(volatile DWORD *)(SLCR + SLCR_MIO_PIN_00_OFFSET + (i * sizeof(DWORD)));
        if (dwRegisterData != MIO_PIN_QSPI_3V_NO_PULLUP_VALUE)
        {
            printf("\n Updating MIO pin %d from %08lxh to %08lxh", i, dwRegisterData,
                                        MIO_PIN_QSPI_3V_NO_PULLUP_VALUE);
            *(volatile DWORD *)(SLCR + SLCR_MIO_PIN_00_OFFSET + (i * sizeof(DWORD))) = MIO_PIN_QSPI_3V_NO_PULLUP_VALUE;
        }
    }

    // Reset the controller
    dwRegisterData = SLCR_QSPI_RST_CTRL_REG;
    dwRegisterData |= (QSPI_REF_RST_VALUE | LQSPI_CPU1X_RST_VALUE);     // set both reset bits
    SLCR_QSPI_RST_CTRL_REG = dwRegisterData;
    dwRegisterData &= (~QSPI_REF_RST_VALUE & ~LQSPI_CPU1X_RST_VALUE);   // clear both reset bits
    SLCR_QSPI_RST_CTRL_REG = dwRegisterData;

    // Make sure we are talking to the right controller
    dwRegisterData = SPI_MOD_ID;
    if (dwRegisterData != SPI_MOD_ID_VALUE)
    {
        printf("\n\n Module ID read as %08lxh, but expected %08lxh", dwRegisterData, SPI_MOD_ID_VALUE);
        return(SLLD_E_HAL_ERROR);
    }

    // Disable any ongoing operation and interrupts
    SPI_EN_REG = SPI_EN_REG_DISABLE_VALUE;
    SPI_INTRPT_DIS_REG = SPI_INTRPT_ALL_VALUE;

#ifdef DUAL_QSPI
    // Disable linear mode, dual flash (parallel)
    SPI_LQSPI_CFG = SPI_LQSPI_CFG_DUAL_FLASH_VALUE;
#else
    // Disable linear mode, single flash
    SPI_LQSPI_CFG = SPI_LQSPI_CFG_DISABLE_VALUE;
#endif

    while (SPI_INTRPT_STATUS_REG & SPI_INTRPT_STATUS_REG_RXNEMPTY_MASK)  // data in RX FIFO?
        dwRegisterData = SPI_RX_DATA_REG;

    SPI_INTRPT_STATUS_REG = SPI_INTRPT_ALL_VALUE;       // clear possible TX FIFO and RX FIFO interrupt flags

    dwRegisterData = SPI_CONFIG_REG;
    dwRegisterData &= (~SPI_CONFIG_REG_ENDIAN_MASK              // little endian
                        & ~SPI_CONFIG_REG_CPHA_MASK             // mode 0, CPHA = 0
                        & ~SPI_CONFIG_REG_CPOL_MASK             // mode 0, CPOL = 0
                        & ~SPI_CONFIG_REG_BAUD_RATE_DIV_MASK);  // clear the clock divisor field (100MHz)
    dwRegisterData |= (SPI_CONFIG_REG_IFMODE_MASK               // support dual/quad mode, not legacy
                        | SPI_CONFIG_REG_MANSTRTEN_MASK         // enable manual start of SPI operations
                        | SPI_CONFIG_REG_SSFORCE_MASK           // manual handling of CS#
                        | SPI_CONFIG_REG_PCS_MASK               // deassert CS# (1)
                        | SPI_CONFIG_REG_FIFO_WIDTH_MASK        // FIFO bus is 32 bits
//                        | SPI_CONFIG_REG_BAUD_RATE_DIV_4        // divide input clock by 4 (50 MHz)
//                        | SPI_CONFIG_REG_BAUD_RATE_DIV_8        // divide input clock by 8 (25 MHz)
                        | SPI_CONFIG_REG_MSTREN_MASK);          // put the SPI device in master mode
    SPI_CONFIG_REG = dwRegisterData;

    // Deactivate WP# signal, although it should be 1 (inacivate) by default
    dwRegisterData = SPI_GPIO;
    dwRegisterData |= SPI_GPIO_WP_N_MASK;
    SPI_GPIO = dwRegisterData;

    // Enable the SPI controller for use
    SPI_EN_REG = SPI_EN_REG_ENABLE_VALUE;

    return(SLLD_OK);
}


// ***************************************************************************
//  FLASH_READ - HAL read function
//
//  input : device_num            device number to which operation will be done
//          command               write a single command byte to flash
//          sys_addr              system address to be used
//          data_buffer           Pointer to the data buffer where to store the read data
//          Number_Of_Read_Bytes  number of bytes to be read
//
//  return value : status of the operation - FAIL or SUCCESS
// ***************************************************************************
SLLD_STATUS FLASH_READ
(
BYTE    device_num,                     /* device number */
BYTE     command,                        /* write a single command byte to flash */
ADDRESS  sys_addr,                       /* system address to be used */
BYTE    *data_buffer,                    /* Pointer to the data buffer containing data to be written */
int      Number_Of_Read_Bytes            /* number of bytes to be read */
)
{
    SLLD_STATUS status = SLLD_OK;
    int         data_byte_index;
    int         junk_word_index;
    int         cmd_addr_bytes_written;
    int         data_bytes;
    int         data_offset;            // offset of valid data within a word
    int         round_up;
    int         current_read_bytes;
    DWORD       dwRegisterData;
    int         tx_fifo_empty;
//    BYTE        bModeBits = 0;      // 0xA? (lower nibble is don't care) to enter high performance read mode

    if ((sys_addr == ADDRESS_NOT_USED) && ((command == SPI_READ_CMD)
                                        || (command == SPI_FAST_READ_CMD)
                                        || (command == SPI_DUALIO_RD_CMD)
                                        || (command == SPI_QUADIO_RD_CMD)
                                        || (command == SPI_READID_90_CMD)
                                        || (command == SPI_DUALIO_HPRD_CMD)
                                        || (command == SPI_DYB_RD_CMD)
                                        || (command == SPI_PPB_RD_CMD)
                                        || (command == SPI_QUADIO_HPRD_CMD)
                                        || (command == SPI_READ_4B_CMD)
                                        || (command == SPI_FAST_READ_4B_CMD)
                                        || (command == SPI_DUALIO_RD_4B_CMD)
                                        || (command == SPI_QUADIO_RD_4B_CMD)
                                        || (command == SPI_DUALIO_HPRD_4B_CMD)
                                        || (command == SPI_QUADIO_HPRD_4B_CMD)))
    {
        printf("\nSPI opcode %02xh is supposed to have address cycles!\n", command);
        return (SLLD_E_HAL_ERROR);
    }

    if ((sys_addr != ADDRESS_NOT_USED) && ((command == SPI_RDSR_CMD)
                                        || (command == SPI_RDSR2_CMD)
                                        || (command == SPI_ABRD_CMD)
                                        || (command == SPI_BRRD_CMD)
                                        || (command == SPI_RASP_CMD)
                                        || (command == SPI_RCR_CMD)
                                        || (command == SPI_RDID_9F_CMD)
                                        || (command == SPI_PASSRD_CMD)
                                        || (command == SPI_RDID_AB_CMD)))
    {
        printf("\nSPI opcode %02xh is not supposed to have address cycles!\n", command);
        return (SLLD_E_HAL_ERROR);
    }

    if ((Number_Of_Read_Bytes != 0) && (!data_buffer))
    {
        printf("\nNULL buffer for storing read data!\n");
        return (SLLD_E_HAL_ERROR);
    }

    // Check for TX FIFO full
    if (SPI_INTRPT_STATUS_REG & SPI_INTRPT_STATUS_REG_TXFULL_MASK)
    {
        printf("\nTX FIFO full upon entry to FLASH_READ!\n");
        return (SLLD_E_HAL_ERROR);
    }

    // For dual SPI, shift the address
#ifdef DUAL_QSPI

#endif

    // Assert chip select
    dwRegisterData = SPI_CONFIG_REG;
    dwRegisterData &= ~SPI_CONFIG_REG_PCS_MASK;     // assert CS# (0)
    SPI_CONFIG_REG = dwRegisterData;

    // Select the TX FIFO register based on the command
    switch(command)
    {
        case SPI_READ_CMD:
        case SPI_READID_90_CMD:
        case SPI_DYB_RD_CMD:
        case SPI_PPB_RD_CMD:
        {
            cmd_addr_bytes_written = 4;
            dwRegisterData = (ADDR_UPPER_3B | command);
            SPI_TXD0 = dwRegisterData;          // one byte opcode, three address
            break;
        }
        case SPI_FAST_READ_CMD:
        case SPI_DUALIO_RD_CMD:
        case SPI_QUADIO_RD_CMD:
        {
            if (sys_addr >= 3)
            {
                sys_addr -= 3;                  // read 3 extra bytes to align RX FIFO with data_buffer
                cmd_addr_bytes_written = 8;     // add 3 more bytes to throw away at the start
            }
            else
                cmd_addr_bytes_written = 5;
            dwRegisterData = (ADDR_UPPER_3B | command);
            SPI_TXD0 = dwRegisterData;          // one byte opcode, three address
            SPI_TXD0 = 0;                       // one dummy byte
            break;
        }
        case SPI_DUALIO_HPRD_CMD:
        {
            if (sys_addr >= 3)
            {
                sys_addr -= 3;                  // read 3 extra bytes to align RX FIFO with data_buffer
                cmd_addr_bytes_written = 8;     // add 3 more bytes to throw away at the start
            }
            else
                cmd_addr_bytes_written = 5;
            dwRegisterData = (ADDR_UPPER_3B | command);
            SPI_TXD0 = dwRegisterData;          // one byte opcode, three address
            SPI_TXD0 = (DWORD)modebit_char;     // one dummy byte is also the mode bits
            break;
        }
        case SPI_QUADIO_HPRD_CMD:
        {
            if (sys_addr >= 1)
            {
                sys_addr -= 1;                  // read 1 extra byte to align RX FIFO with data_buffer
                cmd_addr_bytes_written = 8;     // add 1 more byte to throw away at the start
            }
            else
                cmd_addr_bytes_written = 7;
            dwRegisterData = (ADDR_UPPER_3B | command);
            SPI_TXD0 = dwRegisterData;          // one byte opcode, three address
            SPI_TXD0 = (DWORD)modebit_char;     // three dummy bytes, first byte (LSB) is the mode bits
            break;
        }
        case SPI_RDSR_CMD:
        case SPI_RDSR2_CMD:
        case SPI_ABRD_CMD:
        case SPI_BRRD_CMD:
        case SPI_RASP_CMD:
        case SPI_RCR_CMD:
        case SPI_RDID_9F_CMD:
        //case SPI_READ_ID_9F_CMD:  // same value as SPI_RDID_9F_CMD
        case SPI_PASSRD_CMD:
        {
            cmd_addr_bytes_written = 1;
            SPI_TXD0 = command;                 // one byte opcode, zero address
            break;
        }
        case SPI_RDID_AB_CMD:
        //case SPI_RES_CMD:     // same value as SPI_RDID_AB_CMD
        {
            cmd_addr_bytes_written = 4;
            SPI_TXD0 = (DWORD)command;          // one byte opcode, zero address, three dummy bytes
            break;
        }
        case SPI_READ_4B_CMD:
        {
            sys_addr -= 3;                      // read 3 extra bytes to align RX FIFO with data_buffer
            cmd_addr_bytes_written = 8;         // add 3 more bytes to throw away at the start
            dwRegisterData = (ADDR_UPPER_4B | command);
            SPI_TXD0 = dwRegisterData;          // one byte opcode, three address
            SPI_TXD0 = ADDR_LOWER_4B;           // low address byte
            break;
        }
        case SPI_FAST_READ_4B_CMD:
        case SPI_DUALIO_RD_4B_CMD:
        case SPI_QUADIO_RD_4B_CMD:
        {
            sys_addr -= 2;                      // read 2 extra bytes to align RX FIFO with data_buffer
            cmd_addr_bytes_written = 8;         // add 2 more bytes to throw away at the start
            dwRegisterData = (ADDR_UPPER_4B | command);
            SPI_TXD0 = dwRegisterData;          // one byte opcode, three address
            SPI_TXD0 = ADDR_LOWER_4B;           // one dummy byte and low address byte
            break;
        }
        case SPI_DUALIO_HPRD_4B_CMD:
        {
            sys_addr -= 2;                      // read 2 extra bytes to align RX FIFO with data_buffer
            cmd_addr_bytes_written = 8;         // add 2 more bytes to throw away at the start
            dwRegisterData = (ADDR_UPPER_4B | command);
            SPI_TXD0 = dwRegisterData;          // one byte opcode, three address
            SPI_TXD0 = ((modebit_char << 8) | ADDR_LOWER_4B);   // One dummy byte (mode bits) and low address byte
            break;
        }
        case SPI_QUADIO_HPRD_4B_CMD:
        {
            cmd_addr_bytes_written = 8;
            dwRegisterData = (ADDR_UPPER_4B | command);
            SPI_TXD0 = dwRegisterData;          // one byte opcode, three address
            SPI_TXD0 = ((modebit_char << 8) | ADDR_LOWER_4B);   // Three dummy bytes and low address byte
            break;                                              // mode bits are the first of the dummy bytes
        }
        default:
        {
            status = SLLD_E_HAL_ERROR;
            printf("\nThis SPI opcode is not supported!\n");
            goto EXIT;
        }
    }

    while (Number_Of_Read_Bytes > 0)
    {
        round_up = ((cmd_addr_bytes_written + Number_Of_Read_Bytes) % SPI_CONTROLLER_WORD_SIZE) ? 1 : 0;
        if ((((cmd_addr_bytes_written + Number_Of_Read_Bytes) / SPI_CONTROLLER_WORD_SIZE) + round_up)
                            < (SPI_FIFO_SIZE_IN_BYTES / SPI_CONTROLLER_WORD_SIZE))
            current_read_bytes = Number_Of_Read_Bytes;
        else
            current_read_bytes = SPI_FIFO_SIZE_IN_BYTES - cmd_addr_bytes_written;

        // transmit extra dummy bytes equivalent to the bytes we want to read (specific to the Zynq SPI controller)
        data_byte_index = current_read_bytes;

        // adjust for any dummy bytes already written to the TX FIFO along with command/address
        if (cmd_addr_bytes_written % SPI_CONTROLLER_WORD_SIZE)
            data_byte_index -= (SPI_CONTROLLER_WORD_SIZE - (cmd_addr_bytes_written % SPI_CONTROLLER_WORD_SIZE));

        while (data_byte_index > 0) // count down from the number of bytes to zero
        {
            SPI_TXD0 = 0;
            data_byte_index -= 4;   // we might read an extra 1, 2, or 3 bytes from flash
        }

        dwRegisterData = SPI_CONFIG_REG;
        dwRegisterData |= SPI_CONFIG_REG_MANSTRT_MASK;      // start transmitting SPI sequence to flash
        SPI_CONFIG_REG = dwRegisterData;
        tx_fifo_empty = FALSE;

        // Don't wait for TX FIFO empty, just start reading (garbage and real data)
        data_byte_index = 0;
        junk_word_index = 0;
        while (data_byte_index < current_read_bytes)
        {
            if ((tx_fifo_empty == TRUE) ||
                (SPI_INTRPT_STATUS_REG & SPI_INTRPT_STATUS_REG_RXNEMPTY_MASK)) // data in RX FIFO
            {
                dwRegisterData = SPI_RX_DATA_REG;
                data_offset = 0;    // default to valid data at the start of the word

                if (junk_word_index < (cmd_addr_bytes_written / SPI_CONTROLLER_WORD_SIZE))
                {
                    // throw out this word, it represents writing command/address/dummy cycles to the flash
                    junk_word_index++;
                    data_bytes = 0;
                }
                else
                {
                    if ((cmd_addr_bytes_written % SPI_CONTROLLER_WORD_SIZE) && (data_byte_index == 0))
                    {
                        // this word has junk from cmd/address and real data
                        if (current_read_bytes < (SPI_CONTROLLER_WORD_SIZE -
                                            (cmd_addr_bytes_written % SPI_CONTROLLER_WORD_SIZE)))
                            data_bytes = current_read_bytes;
                        else
                            data_bytes = SPI_CONTROLLER_WORD_SIZE - (cmd_addr_bytes_written % SPI_CONTROLLER_WORD_SIZE);

                        data_offset = (cmd_addr_bytes_written % SPI_CONTROLLER_WORD_SIZE);
                    }
                    else
                        data_bytes = current_read_bytes - data_byte_index;
                }

                switch (data_bytes)
                {
                    case 0:     // don't copy any bytes from this word
                    {
                        break;
                    }
                    case 1:     // only copy one byte
                    {
                        *(data_buffer + data_byte_index) = (BYTE)(dwRegisterData >> SHIFT);    // copy the high byte
                        data_byte_index++;
                        break;
                    }
                    case 2:     // only copy two bytes
                    {
                        *(WORD *)(data_buffer + data_byte_index) = (WORD)(dwRegisterData >> SHIFT);    // upper two bytes
                        data_byte_index += 2;
                        break;
                    }
                    case 3:     // only copy three bytes
                    {
                        *(WORD *)(data_buffer + data_byte_index) = (WORD)(dwRegisterData >> SHIFT);
                        *(data_buffer + data_byte_index + 2) = (BYTE)(dwRegisterData >> (SHIFT + 16));
                        data_byte_index += 3;
                        break;
                    }
                    default:    // 4 or more bytes, copy the entire word
                    {
                        *(DWORD *)(data_buffer + data_byte_index) = dwRegisterData;
                        data_byte_index += 4;
                        break;
                    }
                }
            }

            if (tx_fifo_empty == FALSE)
            {
                dwRegisterData = SPI_INTRPT_STATUS_REG;
                // Check for the TX not full interrupt flag (default threshold is 1, so this indicates TX FIFO is empty)
                if (dwRegisterData & SPI_INTRPT_STATUS_REG_TXOW_MASK)
                {
                    // Now that the transfer is complete, read the remaining bytes without checking for RX not empty
                    tx_fifo_empty = TRUE;
                }
            }
        }

        if (SPI_INTRPT_STATUS_REG & SPI_INTRPT_STATUS_REG_RXOVR_MASK) // data overflow in RX FIFO?
        {
            status = SLLD_E_HAL_ERROR;
            printf("\nExpected to read %d bytes, but RX FIFO overflowed!\n", current_read_bytes);
            goto EXIT;
        }

        while (SPI_INTRPT_STATUS_REG & SPI_INTRPT_STATUS_REG_RXNEMPTY_MASK)   // extra data in RX FIFO
        {
            dwRegisterData = SPI_RX_DATA_REG;   // if we read extra bytes, throw them out
        }

        Number_Of_Read_Bytes -= current_read_bytes; // reduce total data count by current data count
        data_buffer += current_read_bytes; // update the buffer pointer for storing data read
        cmd_addr_bytes_written = 0;     // only used in the first loop, when writing command and address
    }

EXIT:
    // Deassert chip select
    dwRegisterData = SPI_CONFIG_REG;
    dwRegisterData |= SPI_CONFIG_REG_PCS_MASK;      // deassert CS# (1)
    SPI_CONFIG_REG = dwRegisterData;

    return(status);
}


// ***************************************************************************
//  FLASH_WRITE - HAL write function
//
//  input : device_num               device number to which operation will be done
//          command                  write a single command byte to flash
//          sys_addr                 system address to be used
//          data_buffer              optional pointer to the data buffer
//          Number_Of_Written_Bytes  number of bytes to be written
//
//  return value : status of the operation - FAIL or SUCCESS
// ***************************************************************************
SLLD_STATUS FLASH_WRITE
(
BYTE    device_num,                   /* device number */
BYTE     command,                        /* write a single command byte to flash */
ADDRESS  sys_addr,                       /* system address to be used */
BYTE    *data_buffer,                    /* Pointer to the data buffer containing data to be written */
int      Number_Of_Written_Bytes         /* number of bytes to be written */
)
{
    SLLD_STATUS status = SLLD_OK;
    int         data_byte_index;
    int         cmd_addr_bytes_written;
    int         round_up;
    int         current_write_bytes;
    DWORD       dwRegisterData;

    if ((sys_addr == ADDRESS_NOT_USED) && ((command == SPI_PP_CMD)
                                        || (command == SPI_P4E_CMD)
                                        || (command == SPI_QPP_CMD)
                                        || (command == SPI_QPP2_CMD)
                                        || (command == SPI_P8E_CMD)
                                        || (command == SPI_BLOCK_ERASE_CMD) // same as SPI_SE_CMD
                                        || (command == SPI_DYB_PG_CMD)
                                        || (command == SPI_PPB_PG_CMD)
                                        || (command == SPI_PP_4B_CMD)))
    {
        printf("\nSPI opcode %02xh is supposed to have address cycles!\n", command);
        return (SLLD_E_HAL_ERROR);
    }

    if ((sys_addr != ADDRESS_NOT_USED) && ((command == SPI_WRSR_CMD)    // same as SPI_WRR_CMD
                                        || (command == SPI_WRDI_CMD)
                                        || (command == SPI_WREN_CMD)
                                        || (command == SPI_ABWR_CMD)
                                        || (command == SPI_BRWR_CMD)
                                        || (command == SPI_WASP_CMD)
                                        || (command == SPI_CLSR_CMD)
                                        || (command == SPI_BE1_CMD)
                                        || (command == SPI_ERS_SSP_CMD)
                                        || (command == SPI_ERS_RES_CMD)
                                        || (command == SPI_PGSP_CMD)
                                        || (command == SPI_PGRS_CMD)
                                        || (command == SPI_DP_CMD)      // deep power down, BRAC on FL-S
                                        || (command == SPI_BE_CMD)
                                        || (command == SPI_PPB_ERS_CMD)
                                        || (command == SPI_PASSP_CMD)   // same as SPI_WPWD_CMD
                                        || (command == SPI_PASSU_CMD)
                                        || (command == SPI_SOFTWARE_RESET)
                                        || (command == SPI_READMODE_RESET_CMD)))
    {
        printf("\nSPI opcode %02xh is not supposed to have address cycles!\n", command);
        return (SLLD_E_HAL_ERROR);
    }

    if ((Number_Of_Written_Bytes != 0) && (!data_buffer))
    {
        printf("\nNULL buffer for data to program!\n");
        return (SLLD_E_HAL_ERROR);
    }

    // Check for TX FIFO full
    if (SPI_INTRPT_STATUS_REG & SPI_INTRPT_STATUS_REG_TXFULL_MASK)
    {
        printf("\nTX FIFO full upon entry to FLASH_WRITE!\n");
        return (SLLD_E_HAL_ERROR);
    }

    // Assert chip select
    dwRegisterData = SPI_CONFIG_REG;
    dwRegisterData &= ~SPI_CONFIG_REG_PCS_MASK;     // assert CS# (0)
    SPI_CONFIG_REG = dwRegisterData;

    // For dual SPI, shift the address
#ifdef DUAL_QSPI

#endif


    // Select the TX FIFO register based on the command
    switch(command)
    {
        case SPI_PP_CMD:
        case SPI_P4E_CMD:
        case SPI_QPP_CMD:
        case SPI_QPP2_CMD:
        case SPI_P8E_CMD:       // supported on FL-P, but not FL-S
        case SPI_BLOCK_ERASE_CMD:
        //case SPI_SE_CMD:      // same value as SPI_BLOCK_ERASE_CMD
        case SPI_DYB_PG_CMD:
        case SPI_PPB_PG_CMD:
        {
            dwRegisterData = (ADDR_UPPER_3B | command);
            SPI_TXD0 = dwRegisterData;          // one byte opcode, three address
            cmd_addr_bytes_written = 4;
            break;
        }
        case SPI_WRSR_CMD:
        //case SPI_WRR_CMD:     // same value as SPI_WRSR_CMD
        {
            switch(Number_Of_Written_Bytes)
            {
                case 1:
                {
                    SPI_TXD2 = (command | ((*(data_buffer) << 8) & 0x0000FF00));    // one byte opcode, one data
                    cmd_addr_bytes_written = 2; // actually 1 was data
                    break;
                }
                case 2:
                {
                    SPI_TXD3 = (command | ((*(WORD *)(data_buffer) << 8) & 0x00FFFF00));    // one byte opcode, two data
                    cmd_addr_bytes_written = 3; // actually 2 were data
                    break;
                }
                default:
                {
                    status = SLLD_E_HAL_ERROR;
                    printf("\nInvalid to write %d bytes with the WRR command!\n", Number_Of_Written_Bytes);
                    goto EXIT;
                }
            Number_Of_Written_Bytes = 0;        // don't write any more data
            break;
            }
        }
        case SPI_WRDI_CMD:
        case SPI_WREN_CMD:
        case SPI_CLSR_CMD:
        case SPI_BE1_CMD:
        case SPI_ERS_SSP_CMD:
        case SPI_ERS_RES_CMD:
        case SPI_PGSP_CMD:
        case SPI_PGRS_CMD:
        case SPI_DP_CMD:        // deep power down, BRAC on FL-S
        case SPI_BE_CMD:
        case SPI_PPB_ERS_CMD:
        case SPI_SOFTWARE_RESET:
        case SPI_READMODE_RESET_CMD:
        {
            SPI_TXD1 = command;             // one byte opcode, zero address, zero data
            cmd_addr_bytes_written = 1;
            break;
        }
        case SPI_ABWR_CMD:
        case SPI_BRWR_CMD:
        {
            SPI_TXD2 = (command | ((*(data_buffer) << 8) & 0x0000FF00));    // one byte opcode, one data
            cmd_addr_bytes_written = 2;         // actually 1 was data
            Number_Of_Written_Bytes = 0;        // don't write any more data
            break;
        }
        case SPI_WASP_CMD:
        {
            SPI_TXD3 = (command | ((*(WORD *)(data_buffer) << 8) & 0x00FFFF00));    // one byte opcode, two data
            cmd_addr_bytes_written = 3;         // actually 2 were data
            Number_Of_Written_Bytes = 0;        // don't write any more data
            break;
        }
        case SPI_PASSP_CMD:
        //case SPI_WPWD_CMD:    // same value as SPI_PASSP_CMD
        case SPI_PASSU_CMD:
        {
            dwRegisterData = (command | ((*(data_buffer) << 8) & 0x0000FF00));
            dwRegisterData |= (((*(WORD *)(data_buffer + 1)) << 16) & 0xFFFF0000);
            SPI_TXD0 = dwRegisterData;              // one byte opcode, three data
            SPI_TXD0 = *(DWORD *)(data_buffer + 3); // four more data bytes
            FLASH_EMPTY_TXFIFO();
            SPI_TXD1 = *(data_buffer + 7);          // one more data byte
            cmd_addr_bytes_written = 9;         // actuallly 8 were data
            Number_Of_Written_Bytes = 0;        // don't write any more data
            break;
        }
        case SPI_PP_4B_CMD:
        case SPI_QPP_4B_CMD:
        {
            dwRegisterData = (ADDR_UPPER_4B | command);
            SPI_TXD0 = dwRegisterData;          // one byte opcode, three address

            switch(Number_Of_Written_Bytes)
            {
                case 0:
                {
                    FLASH_EMPTY_TXFIFO();
                    dwRegisterData = SPI_RX_DATA_REG;   // read one word from the RX FIFO
                    status = SLLD_E_HAL_ERROR;
                    printf("\nInvalid to write zero bytes with the %02xh program command!\n", command);
                    goto EXIT;
                }
                case 1:
                {
                    FLASH_EMPTY_TXFIFO();
                    SPI_TXD2 = (ADDR_LOWER_4B | ((*(data_buffer) << 8) & 0x0000FF00));  // one byte address, one data
                    cmd_addr_bytes_written = 6;         // actually 1 was data
                    Number_Of_Written_Bytes = 0;        // don't write any more data
                    break;
                }
                case 2:
                {
                    FLASH_EMPTY_TXFIFO();
                    SPI_TXD3 = (ADDR_LOWER_4B | ((*(WORD *)(data_buffer) << 8) & 0x00FFFF00));  // one byte opcode, two data
                    cmd_addr_bytes_written = 7;         // actually 2 were data
                    Number_Of_Written_Bytes = 0;        // don't write any more data
                    break;
                }
                default:    // 3 or more data bytes
                {
                    SPI_TXD0 = (ADDR_LOWER_4B | ((*(WORD *)(data_buffer) << 8) & 0xFFFFFF00));    // one byte opcode, two data
                    cmd_addr_bytes_written = 8;         // actually 3 were data
                    Number_Of_Written_Bytes -= 3;       // already wrote the first three data bytes
                    break;
                }
            break;
            }
        }
        case SPI_P8E_4B_CMD:
        case SPI_SE_4B_CMD:
        {
            dwRegisterData = (ADDR_UPPER_4B | command);
            SPI_TXD0 = dwRegisterData;          // one byte opcode, three address
            FLASH_EMPTY_TXFIFO();
            SPI_TXD1 = ADDR_LOWER_4B;           // low address byte
            cmd_addr_bytes_written = 5;
            break;
        }
        default:
        {
            status = SLLD_E_HAL_ERROR;
            printf("\nThis SPI opcode is not supported!\n");
            goto EXIT;
        }
    }

    do
    {
        round_up = ((cmd_addr_bytes_written + Number_Of_Written_Bytes) % SPI_CONTROLLER_WORD_SIZE) ? 1 : 0;
        if ((((cmd_addr_bytes_written + Number_Of_Written_Bytes) / SPI_CONTROLLER_WORD_SIZE) + round_up)
                                < (SPI_FIFO_SIZE_IN_BYTES / SPI_CONTROLLER_WORD_SIZE))
            current_write_bytes = Number_Of_Written_Bytes;
        else
            current_write_bytes = SPI_FIFO_SIZE_IN_BYTES - cmd_addr_bytes_written;

        // Write the data
        data_byte_index = 0;
        while (current_write_bytes > data_byte_index)
        {
            if (SPI_INTRPT_STATUS_REG & SPI_INTRPT_STATUS_REG_TXFULL_MASK)    // no room in TX FIFO
            {
                status = SLLD_E_HAL_ERROR;
                printf("\nMore data to write and TX FIFO is full!\n");
                goto EXIT;
            }

            switch (current_write_bytes - data_byte_index)
            {
                case 1:
                {
                    FLASH_EMPTY_TXFIFO();
                    SPI_TXD1 = *(data_buffer + data_byte_index);
                    data_byte_index++;
                    break;
                }
                case 2:
                {
                    FLASH_EMPTY_TXFIFO();
                    SPI_TXD2 = *(WORD *)(data_buffer + data_byte_index);
                    data_byte_index += 2;
                    break;
                }
                case 3:
                {
                    FLASH_EMPTY_TXFIFO();
                    dwRegisterData = (*(WORD *)(data_buffer + data_byte_index) & 0x0000FFFF);
                    dwRegisterData |= (((*(data_buffer + data_byte_index + 2)) << 16) & 0x00FF0000);
                    SPI_TXD3 = dwRegisterData;
                    data_byte_index += 3;
                    break;
                }
                default:    // 4 or more
                {
                    SPI_TXD0 = *(DWORD *)(data_buffer + data_byte_index);
                    data_byte_index += 4;
                    break;
                }
            }

            if (current_write_bytes < data_byte_index)
            {
                status = SLLD_E_HAL_ERROR;
                printf("\nWrote too many bytes!\n");
                goto EXIT;
            }
        }

        // Did read FIFO interrupts allow the right number of data writes?
        if (current_write_bytes != data_byte_index)
        {
            status = SLLD_E_HAL_ERROR;
            printf("\nWrote wrong number of bytes!\n");
            goto EXIT;
        }

        dwRegisterData = SPI_CONFIG_REG;
        dwRegisterData |= SPI_CONFIG_REG_MANSTRT_MASK;      // start transmitting SPI command sequence to flash
        SPI_CONFIG_REG = dwRegisterData;

        dwRegisterData = FLASH_POLL();                      // wait for TX FIFO not full interrupt

        if (!(dwRegisterData & SPI_INTRPT_STATUS_REG_TXOW_MASK))
        {
            status = SLLD_E_HAL_ERROR;
            printf("\nFLASH_POLL failed waiting for TX FIFO not full!  Interrupt reg value = %02xh\n", (unsigned int)dwRegisterData);
            goto EXIT;
        }

        // Read the data (all garbage)
        for (data_byte_index = 0; data_byte_index < (current_write_bytes
                        + cmd_addr_bytes_written); data_byte_index += 4)
        {
            dwRegisterData = SPI_RX_DATA_REG;
        }

        // Any remaining data in RX FIFO?
        while (SPI_INTRPT_STATUS_REG & SPI_INTRPT_STATUS_REG_RXNEMPTY_MASK)
        {
            dwRegisterData = SPI_RX_DATA_REG;
        }

        Number_Of_Written_Bytes -= current_write_bytes; // reduce total data count by current data count
        data_buffer += current_write_bytes; // update the buffer pointer for data to write
        cmd_addr_bytes_written = 0;     // only used in the first loop, when writing command and address
    }while (Number_Of_Written_Bytes > 0);

EXIT:
    // Deassert chip select
    dwRegisterData = SPI_CONFIG_REG;
    dwRegisterData |= SPI_CONFIG_REG_PCS_MASK;      // deassert CS# (1)
    SPI_CONFIG_REG = dwRegisterData;

    return(status);
}


// ***************************************************************************
//  FLASH_POLL - HAL polling function
//
//  input : none
//
//  return value : value of the QSPI interrupt register, before clearing
// ***************************************************************************
DWORD FLASH_POLL
(
void
)
{
    int         polling_index;
    DWORD       dwRegisterData;

    // Gary Test - LED9 off to indicate polling started
    GPIO_MASK_DATA0_LSW_REG = GPIO_MIO7_OUTPUT_0;

    for (polling_index = 0; polling_index < SPI_INTRPT_MAX_POLLS; polling_index++)
    {
        dwRegisterData = SPI_INTRPT_STATUS_REG;

        // Poll for the TX not full interrupt flag (default threshold is 1, so this indicates TX FIFO is empty)
        if (dwRegisterData & SPI_INTRPT_STATUS_REG_TXOW_MASK)
            break;
    }

    if (polling_index < SPI_INTRPT_MAX_POLLS)
        SPI_INTRPT_STATUS_REG = dwRegisterData;     // clear the interrupt(s)

    // Gary Test - LED9 on to indicate done polling
    GPIO_MASK_DATA0_LSW_REG = GPIO_MIO7_OUTPUT_1;

    return (dwRegisterData);
}


// ***************************************************************************
//  FLASH_EMPTY_TXFIFO - HAL function to empty TX FIFO
//
//  input : none
//
//  return value : none
// ***************************************************************************
void FLASH_EMPTY_TXFIFO
(
void
)
{
    DWORD       dwRegisterData;

    dwRegisterData = SPI_CONFIG_REG;
    dwRegisterData |= SPI_CONFIG_REG_MANSTRT_MASK;      // start transmitting SPI command sequence to flash
    SPI_CONFIG_REG = dwRegisterData;

    dwRegisterData = FLASH_POLL();      // wait for TX FIFO not full interrupt
    if (!(dwRegisterData & SPI_INTRPT_STATUS_REG_TXOW_MASK))
        printf("\nFLASH_POLL failed waiting for TX FIFO not full!  Interrupt reg value = %02xh\n", (unsigned int)dwRegisterData);
}


// ***************************************************************************
//  DelayMicroseconds - Performs a delay.
//
//  input : number of microseconds to delay the processor
//
//  return value : none
// ***************************************************************************
#define DELAY_1us 150
void DelayMicroseconds(int microseconds)
{
    int volatile i, j;

    for (j = 0; j < microseconds; j++)
        for(i = 0; i < DELAY_1us; i++) i = i;

    return;
}

/*****************************************************************************/
