/* slld.c - Source Code for Cypress SPI Flash's Low Level Driver */

/**************************************************************************
 * 2003-2016, Cypress Semiconductor Corporation or a subsidiary of Cypress 
 * Semiconductor Corporation.  All rights reserved.
 * 
 * This software, including source code, documentation and related materials 
 * (Software), is owned by Cypress Semiconductor Corporation or one of its 
 * subsidiaries (Cypress) and is protected by and subject to worldwide patent 
 * protection (United States and foreign), United States copyright laws and 
 * international treaty provisions. Therefore, you may use this Software only 
 * as provided in the license agreement accompanying the software package from 
 * which you obtained this Software (EULA).
 * 
 * If no EULA applies, Cypress hereby grants you a personal, non-exclusive, 
 * non-transferable license to copy, modify, and compile the Software source code 
 * solely for use in connection with Cypress's integrated circuit products.  Any 
 * reproduction, modification, translation, compilation, or representation of 
 * this Software except as specified above is prohibited without the express 
 * written permission of Cypress.
 * 
 * Disclaimer: THIS SOFTWARE IS PROVIDED AS-IS, WITH NO WARRANTY OF ANY KIND, 
 * EXPRESS OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, NONINFRINGEMENT, IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. Cypress 
 * reserves the right to make changes to the Software without notice. Cypress does 
 * not assume any liability arising out of the application or use of the Software 
 * or any product or circuit described in the Software. Cypress does not authorize 
 * its products for use in any products where a malfunction or failure of the Cypress 
 * product may reasonably be expected to result in significant property damage, injury 
 * or death (High Risk Product). By including Cypress's product in a High Risk Product, 
 * the manufacturer of such system or application assumes all risk of such use and in 
 * doing so agrees to indemnify Cypress against all liability.
 */

#include <stdio.h>


#include "slld.h"
#include "slld_hal.h"

#ifdef TRACE
#include "trace.h"
#endif

#ifdef INCLUDE_SOFTWARE_PROTECT_STATUS_CHECK
/* static variable to store SPI flash software protect status */
static DEV_SOFTWARE_PROTECT_STATUS sys_software_protect_status = FLASH_SOFTWARE_UNPROTECTED;
#endif

BYTE modebit_char;
BYTECOUNT g_PAGE_SIZE=PAGE_SIZE;











/******************************************************************************
 *
 * slld_Read_IDCmd - Read ID from SPI Flash
 *
 * This function issues the Read_ID command to SPI Flash and reads out the ID.
 * This command sets the target device in software-unprotected state
 * so this function also sets sys_software_protect_status to FLASH_SOFTWARE_UNPROTECTED.
 *
 * RETURNS: SLLD_OK or SLLD_E_HAL_ERROR
 *
 */
SLLD_STATUS slld_Read_IDCmd
(
        BYTE      device_num,                     //device number
        BYTE    *target                      /* variable in which to store read data */
)
{
    SLLD_STATUS status = SLLD_OK;

    status = FLASH_RD(device_num,SPI_READ_ID_CMD, (ADDRESS)ADDRESS_NOT_USED, target, 1);

#ifdef INCLUDE_SOFTWARE_PROTECT_STATUS_CHECK
    sys_software_protect_status = FLASH_SOFTWARE_UNPROTECTED;
#endif

    return(status);
}

/******************************************************************************
 *
 * slld_RDIDCmd - Read Edentification from SPI Flash
 *
 * This function issues the RDID command to SPI Flash and reads out the ID.
 * On some devices this command is documented in the data sheet as
 * Release from deep power down and read electronic signature
 *
 * RETURNS: SLLD_OK or SLLD_E_HAL_ERROR
 *
 */
SLLD_STATUS slld_RDIDCmd
(
        BYTE      device_num,                     //device number
        BYTE      *target,              /* variable in which to store read data */
        BYTECOUNT  len_in_bytes         /* number of bytes to read */
)
{
    SLLD_STATUS status = SLLD_OK;

#ifdef INCLUDE_SOFTWARE_PROTECT_STATUS_CHECK
    /* check if target device is software protected */
    if(sys_software_protect_status != FLASH_SOFTWARE_UNPROTECTED)
    {
        status = SLLD_E_DEVICE_SOFTWARE_PROTECTED;
        return(status);
    }
#endif

    status = FLASH_RD(device_num,SPI_RDID_CMD, (ADDRESS)ADDRESS_NOT_USED, target, len_in_bytes);

    return(status);
}

/******************************************************************************
 *
 * slld_Read_IdentificationCmd - Read ID from SPI Flash
 *
 * This function issues the Read_ID command to SPI Flash and reads out the ID.
 *
 * RETURNS: SLLD_OK, SLLD_E_DEVICE_SOFTWARE_PROTECTED or SLLD_E_HAL_ERROR
 *
 */
SLLD_STATUS slld_Read_IdentificationCmd
(
        BYTE      device_num,                     //device number
        BYTE    *target,                     /* variable in which to store read data */
        ADDRESS  addr                        /* address offset for the command */
)
{
    SLLD_STATUS status = SLLD_OK;
    ADDRESS    tmp_addr;

#ifdef INCLUDE_SOFTWARE_PROTECT_STATUS_CHECK
    /* check if target device is software protected */
    if(sys_software_protect_status != FLASH_SOFTWARE_UNPROTECTED)
    {
        status = SLLD_E_DEVICE_SOFTWARE_PROTECTED;
        return(status);
    }
#endif
    if (addr < 0x2)
        tmp_addr = addr;
    else
        tmp_addr = (ADDRESS)0;

    status = FLASH_RD(device_num,SPI_READ_IDENTIFICATION_CMD, tmp_addr, target, 2);

    return(status);
}

/******************************************************************************
 *
 * slld_RDSRCmd - Read from Status Register
 *
 * This function issues the RDSR command to SPI Flash and reads from status register.
 *
 * RETURNS: SLLD_OK, SLLD_E_DEVICE_SOFTWARE_PROTECTED or SLLD_E_HAL_ERROR
 *
 */
SLLD_STATUS slld_RDSRCmd
(
        BYTE      device_num,                     //device number
        BYTE    *target                        /* variable in which to store read data */
)
{
    SLLD_STATUS status = SLLD_OK;

#ifdef INCLUDE_SOFTWARE_PROTECT_STATUS_CHECK
    /* check if target device is software protected */
    if(sys_software_protect_status != FLASH_SOFTWARE_UNPROTECTED)
    {
        status = SLLD_E_DEVICE_SOFTWARE_PROTECTED;
        return(status);
    }
#endif
    status = FLASH_RD(device_num,SPI_RDSR_CMD, (ADDRESS)ADDRESS_NOT_USED, target, 1);

    return(status);
}






/******************************************************************************
 *
 * slld_WRENCmd - Issue the write enable command
 *
 * This function issues the WREN command to the SPI Flash.
 *
 * RETURNS: SLLD_OK, SLLD_E_DEVICE_SOFTWARE_PROTECTED or SLLD_E_HAL_ERROR
 *
 */
SLLD_STATUS slld_WRENCmd
(
    BYTE      device_num                     //device number
)
{
    SLLD_STATUS status = SLLD_OK;

#ifdef INCLUDE_SOFTWARE_PROTECT_STATUS_CHECK
    /* check if target device is software protected */
    if(sys_software_protect_status != FLASH_SOFTWARE_UNPROTECTED)
    {
        status = SLLD_E_DEVICE_SOFTWARE_PROTECTED;
        return(status);
    }
#endif

    status = FLASH_WR(device_num,SPI_WREN_CMD, (ADDRESS)ADDRESS_NOT_USED, BUFFER_NOT_USED, 0);

    return(status);
}

/******************************************************************************
 *
 * slld_WRDICmd - Issue the write disable command
 *
 * This function issues the WRDI command to the SPI Flash.
 *
 * RETURNS: SLLD_OK, SLLD_E_DEVICE_SOFTWARE_PROTECTED or SLLD_E_HAL_ERROR
 *
 */
SLLD_STATUS slld_WRDICmd
(
    BYTE      device_num                     //device number
)
{
    SLLD_STATUS status = SLLD_OK;

#ifdef INCLUDE_SOFTWARE_PROTECT_STATUS_CHECK
    /* check if target device is software protected */
    if(sys_software_protect_status != FLASH_SOFTWARE_UNPROTECTED)
    {
        status = SLLD_E_DEVICE_SOFTWARE_PROTECTED;
        return(status);
    }
#endif

    status = FLASH_WR(device_num,SPI_WRDI_CMD, (ADDRESS)ADDRESS_NOT_USED, BUFFER_NOT_USED, 0);

    return(status);
}

/******************************************************************************
 *
 * slld_WRSRCmd - Write to Status Register
 *
 * This function issues a write Status Register command to SPI Flash.
 *
 * RETURNS: SLLD_OK, SLLD_E_DEVICE_SOFTWARE_PROTECTED or SLLD_E_HAL_ERROR
 *
 */
SLLD_STATUS slld_WRSRCmd
(
        BYTE      device_num,                     //device number
        BYTE   *data_buf             /* variable containing data to program */
)
{
    SLLD_STATUS status = SLLD_OK;

#ifdef INCLUDE_SOFTWARE_PROTECT_STATUS_CHECK
    /* check if target device is software protected */
    if(sys_software_protect_status != FLASH_SOFTWARE_UNPROTECTED)
    {
        status = SLLD_E_DEVICE_SOFTWARE_PROTECTED;
        return(status);
    }
#endif

    status = FLASH_WR(device_num,SPI_WRSR_CMD, (ADDRESS)ADDRESS_NOT_USED, data_buf, 1);

    return(status);
}

/******************************************************************************
 *
 * slld_WRSROp - Write Status Register Operation
 *
 * Function writes data to Status Register.
 * Function issues all required commands and polls for completion.
 *
 * RETURNS: SLLD_OK, SLLD_E_DEVICE_SOFTWARE_PROTECTED or SLLD_E_HAL_ERROR
 *
 */
SLLD_STATUS slld_WRSROp
(
        BYTE  device_num,            /* device number */
        BYTE       *data_buf,                 /* variable containing data to program */
        DEVSTATUS  *dev_status_ptr            /* variable to store device status */
)
{
    SLLD_STATUS status = SLLD_OK;

    status = slld_WRENCmd(device_num);
    if(status != SLLD_OK)
        return(status);

    status = slld_WRSRCmd(device_num, data_buf);
    if(status != SLLD_OK)
        return(status);

    status = slld_Poll(device_num, dev_status_ptr);
    if(status != SLLD_OK)
        return(status);

    status = slld_WRDICmd(device_num);          /* just in case WRSROp is operated on protected area */

    return(status);
}

void slld_VersionCmd(void)
{       
    printf("\n%s\n", SLLD_VERSION);
}









/******************************************************************************
 *
 * slld_DPCmd - Deep Power-down
 *
 * This function issues the Deep Power-down command to SPI Flash.
 * This command sets the target device in deep power-down state to reduce power consumption
 *
 * RETURNS: SLLD_OK, SLLD_E_DEVICE_SOFTWARE_PROTECTED or SLLD_E_HAL_ERROR
 *
 */
SLLD_STATUS slld_DPCmd 
(
        BYTE      device_num                     //device number
)
{
    SLLD_STATUS status = SLLD_OK;

#ifdef INCLUDE_SOFTWARE_PROTECT_STATUS_CHECK
    /* check if target device is software protected */
    if(sys_software_protect_status != FLASH_SOFTWARE_UNPROTECTED)
    {
        status = SLLD_E_DEVICE_SOFTWARE_PROTECTED;
        return(status);
    }
#endif

    status = FLASH_WR(device_num,SPI_DP_CMD, ADDRESS_NOT_USED, BUFFER_NOT_USED, 0);

#ifdef INCLUDE_SOFTWARE_PROTECT_STATUS_CHECK
    sys_software_protect_status = FLASH_SOFTWARE_PROTECTED;
#endif

    return(status);
}

/******************************************************************************
 *
 * slld_BRACCmd - Bank Register Acess
 *
 * This function issues the Bank Register Acess command to SPI Flash.
 *
 * RETURNS: SLLD_OK, SLLD_E_DEVICE_SOFTWARE_PROTECTED or SLLD_E_HAL_ERROR
 *
 */
SLLD_STATUS slld_BRACCmd 
(        
        BYTE      device_num                     //device number
)
{
    SLLD_STATUS status = SLLD_OK;

#ifdef INCLUDE_SOFTWARE_PROTECT_STATUS_CHECK
    /* check if target device is software protected */
    if(sys_software_protect_status != FLASH_SOFTWARE_UNPROTECTED)
    {
        status = SLLD_E_DEVICE_SOFTWARE_PROTECTED;
        return(status);
    }
#endif

    status = FLASH_WR(device_num,SPI_BRAC_CMD, ADDRESS_NOT_USED, BUFFER_NOT_USED, 0);

    return(status);
}


/******************************************************************************
 *
 * slld_RESCmd - Release from Software Protect (Deep power-down)
 *
 * This function issues the Release from Software Protect command to SPI Flash.
 * This command sets the target device in software-unprotected state
 * so this function also sets sys_software_protect_status[device_num] to FLASH_SOFTWARE_UNPROTECTED.
 *
 * RETURNS: SLLD_OK or SLLD_E_HAL_ERROR
 *
 */
SLLD_STATUS slld_RESCmd (BYTE device_num)
{
    SLLD_STATUS status = SLLD_OK;

    status = FLASH_WR(device_num,SPI_RES_CMD, ADDRESS_NOT_USED, BUFFER_NOT_USED, 0);

#ifdef INCLUDE_SOFTWARE_PROTECT_STATUS_CHECK
    sys_software_protect_status = FLASH_SOFTWARE_UNPROTECTED;
#endif

    return(status);
}











/******************************************************************************
 *
 * slld_ReadCmd - Read from SPI Flash
 *
 * This function issues the Read command to SPI Flash and reads from the array.
 *
 * RETURNS: SLLD_OK, SLLD_E_DEVICE_SOFTWARE_PROTECTED or SLLD_E_HAL_ERROR
 *
 */
SLLD_STATUS slld_ReadCmd
(
        BYTE      device_num,                     //device number
        ADDRESS   sys_addr,          /* device address given by system */
        BYTE     *target,            /* variable in which to store read data */
        BYTECOUNT len_in_bytes       /* number of bytes to read */
)
{
    SLLD_STATUS status = SLLD_OK;

#ifdef INCLUDE_SOFTWARE_PROTECT_STATUS_CHECK
    /* check if target device is software protected */
    if(sys_software_protect_status != FLASH_SOFTWARE_UNPROTECTED)
    {
        status = SLLD_E_DEVICE_SOFTWARE_PROTECTED;
        return(status);
    }
#endif

    status = FLASH_RD(device_num,SPI_READ_CMD, sys_addr, target, len_in_bytes);

    return(status);
}



/******************************************************************************
 *
 * slld_Fast_ReadCmd - Fast Read from SPI Flash
 *
 * This function issues the Fast Read command to SPI Flash and reads from the array.
 *
 * RETURNS: SLLD_OK, SLLD_E_DEVICE_SOFTWARE_PROTECTED or SLLD_E_HAL_ERROR
 *
 */
SLLD_STATUS slld_Fast_ReadCmd
(
        BYTE      device_num,                     //device number
        ADDRESS    sys_addr,              /* device address given by system */
        BYTE      *target,                /* variable in which to store read data */
        BYTECOUNT  len_in_bytes           /* number of bytes on which to operate */
)
{
    SLLD_STATUS         status = SLLD_OK;

#ifdef INCLUDE_SOFTWARE_PROTECT_STATUS_CHECK
    /* check if target device is software protected */
    if(sys_software_protect_status != FLASH_SOFTWARE_UNPROTECTED)
    {
        status = SLLD_E_DEVICE_SOFTWARE_PROTECTED;
        return(status);
    }
#endif

    status = FLASH_RD(device_num,SPI_FAST_READ_CMD, sys_addr, target, len_in_bytes);

    return(status);
}



























/******************************************************************************
 *
 * slld_BECmd - Bulk Erase ,Chip erase for FL1K
 *
 * This function issues the Bulk Erase command to SPI Flash.
 *
 * RETURNS: SLLD_OK, SLLD_E_DEVICE_SOFTWARE_PROTECTED or SLLD_E_HAL_ERROR
 *
 */
SLLD_STATUS slld_BECmd 
(
        BYTE      device_num                     //device number
)
{
    SLLD_STATUS status = SLLD_OK;

#ifdef INCLUDE_SOFTWARE_PROTECT_STATUS_CHECK
    /* check if target device is software protected */
    if(sys_software_protect_status != FLASH_SOFTWARE_UNPROTECTED)
    {
        status = SLLD_E_DEVICE_SOFTWARE_PROTECTED;
        return(status);
    }
#endif

    status = FLASH_WR(device_num,SPI_BULKERASE_CMD, ADDRESS_NOT_USED, BUFFER_NOT_USED, 0);

    return(status);
}

/******************************************************************************
 *
 * slld_BEOp - Performs a Bulk Erase Operation ,Chip erase for FL1K
 *
 * Function erase all data in a device.
 * Function issues all required commands and polls for completion.
 *
 * RETURNS: SLLD_OK, SLLD_E_DEVICE_SOFTWARE_PROTECTED or SLLD_E_HAL_ERROR
 *
 */
SLLD_STATUS slld_BEOp
(
        BYTE  device_num,            /* device number */
        DEVSTATUS  *dev_status_ptr            /* variable to store device status */
)
{
    SLLD_STATUS status = SLLD_OK;

    status = slld_WRENCmd(device_num);
    if(status != SLLD_OK)
        return(status);

    status = slld_BECmd(device_num);
    if(status != SLLD_OK)
        return(status);

    status = slld_Poll(device_num, dev_status_ptr);
    if(status != SLLD_OK)
        return(status);

    status = slld_WRDICmd(device_num);       /* just in case BEOp is operated on protected area */

    return(status);
}

SLLD_STATUS slld_BE1Cmd 
(
        BYTE      device_num                     //device number
)
{
    SLLD_STATUS status = SLLD_OK;

#ifdef INCLUDE_SOFTWARE_PROTECT_STATUS_CHECK
    /* check if target device is software protected */
    if(sys_software_protect_status != FLASH_SOFTWARE_UNPROTECTED)
    {
        status = SLLD_E_DEVICE_SOFTWARE_PROTECTED;
        return(status);
    }
#endif

    status = FLASH_WR(device_num,SPI_BULKERASE1_CMD, ADDRESS_NOT_USED, BUFFER_NOT_USED, 0);

    return(status);
}

SLLD_STATUS slld_BE1Op
(
        BYTE  device_num,            /* device number */
        DEVSTATUS  *dev_status_ptr       /* variable to store device status */
)
{
    SLLD_STATUS status = SLLD_OK;

    status = slld_WRENCmd(device_num);
    if(status != SLLD_OK)
        return(status);

    status = slld_BE1Cmd(device_num);
    if(status != SLLD_OK)
        return(status);

    status = slld_Poll(device_num, dev_status_ptr);
    if(status != SLLD_OK)
        return(status);

    status = slld_WRDICmd(device_num);       /* just in case BEOp is operated on protected area */

    return(status);
}

/******************************************************************************
 *
 * slld_SECmd - Sector Erase
 *
 * This function issues the Sector Erase command to SPI Flash.
 *
 * RETURNS: SLLD_OK, SLLD_E_DEVICE_SOFTWARE_PROTECTED or SLLD_E_HAL_ERROR
 *
 */
SLLD_STATUS slld_SECmd
(
        BYTE      device_num,                     //device number
        ADDRESS  sys_addr                     /* device address given by system */
)
{
    SLLD_STATUS status = SLLD_OK;

#ifdef INCLUDE_SOFTWARE_PROTECT_STATUS_CHECK
    /* check if target device is software protected */
    if(sys_software_protect_status != FLASH_SOFTWARE_UNPROTECTED)
    {
        status = SLLD_E_DEVICE_SOFTWARE_PROTECTED;
        return(status);
    }
#endif

    status = FLASH_WR(device_num,SPI_SE_CMD, sys_addr, BUFFER_NOT_USED, 0);

    return(status);
}

/******************************************************************************
 *
 * slld_SEOp - Performs a Sector Erase Operation
 *
 * Function erases specified sector.
 * Function issues all required commands and polls for completion.
 *
 * RETURNS: SLLD_OK, SLLD_E_DEVICE_SOFTWARE_PROTECTED or SLLD_E_HAL_ERROR
 *
 */
SLLD_STATUS slld_SEOp
(
        BYTE  device_num,            /* device number */
        ADDRESS     sys_addr,                 /* device address given by system */
        DEVSTATUS  *dev_status_ptr            /* variable to store device status */
)
{
    SLLD_STATUS status = SLLD_OK;

    status = slld_WRENCmd(device_num);
    if(status != SLLD_OK)
        return(status);

#ifdef USE_4B_ADDR_CMDS
    status = slld_SE_4BCmd(device_num, sys_addr);
#else
    status = slld_SECmd(device_num, sys_addr);
#endif
    if(status != SLLD_OK)
        return(status);

    status = slld_Poll(device_num, dev_status_ptr);
    if(status != SLLD_OK)
        return(status);

    status = slld_WRDICmd(device_num);        /* just in case SEOp is operated on a protected area */

    return(status);
}







/******************************************************************************
 *
 * slld_BECmd - BLOCK Erase
 *
 * This function issues the BLOCK Erase command to SPI Flash. The Block Erase
 * command sets all bits in the addressed 64 KB block to 1 (all bytes are FFh).
 *
 * RETURNS: SLLD_OK, SLLD_E_DEVICE_SOFTWARE_PROTECTED or SLLD_E_HAL_ERROR
 *
 */
SLLD_STATUS slld_BlockEraseCmd 
(
        BYTE      device_num,                     //device number
        ADDRESS  sys_addr
)
{
    SLLD_STATUS status = SLLD_OK;

#ifdef INCLUDE_SOFTWARE_PROTECT_STATUS_CHECK
    /* check if target device is software protected */
    if(sys_software_protect_status != FLASH_SOFTWARE_UNPROTECTED)
    {
        status = SLLD_E_DEVICE_SOFTWARE_PROTECTED;
        return(status);
    }
#endif

    status = FLASH_WR(device_num,SPI_BLOCK_ERASE_CMD, sys_addr, BUFFER_NOT_USED, 0);

    return(status);
}

/******************************************************************************
 *
 * slld_BEOp - Performs a Bulk Erase Operation.
 *
 * Function erase all data in a device.
 * Function issues all required commands and polls for completion.
 *
 * RETURNS: SLLD_OK, SLLD_E_DEVICE_SOFTWARE_PROTECTED or SLLD_E_HAL_ERROR
 *
 */
SLLD_STATUS slld_BlockEraseOp
(
        BYTE  device_num,            /* device number */
        ADDRESS     sys_addr,             /* device address given by system */
        DEVSTATUS  *dev_status_ptr       /* variable to store device status */
)
{
    SLLD_STATUS status = SLLD_OK;

    status = slld_WRENCmd(device_num);
    if(status != SLLD_OK)
        return(status);

    status = slld_BlockEraseCmd(device_num, sys_addr);
    if(status != SLLD_OK)
        return(status);

    status = slld_Poll(device_num, dev_status_ptr);
    if(status != SLLD_OK)
        return(status);

    status = slld_WRDICmd(device_num);       /* just in case BEOp is operated on protected area */

    return(status);
}

/******************************************************************************
 *
 * slld_PPCmd - Page Program
 *
 * This function issues a Page Program command to SPI Flash.
 *
 * RETURNS: SLLD_OK, SLLD_E_DEVICE_SOFTWARE_PROTECTED or SLLD_E_HAL_ERROR
 *
 */
SLLD_STATUS slld_PPCmd
(
        BYTE      device_num,                     //device number
        ADDRESS     sys_addr,                  /* device address given by system */
        BYTE       *data_buf,                  /* variable containing data to program */
        BYTECOUNT   len_in_bytes               /* number of bytes to operate */
)
{
    SLLD_STATUS  status = SLLD_OK;

#ifdef INCLUDE_SOFTWARE_PROTECT_STATUS_CHECK
    /* check if target device is software protected */
    if(sys_software_protect_status != FLASH_SOFTWARE_UNPROTECTED)
    {
        status = SLLD_E_DEVICE_SOFTWARE_PROTECTED;
        return(status);
    }
#endif

    status = FLASH_WR(device_num,SPI_PP_CMD, sys_addr, data_buf, len_in_bytes);

    return(status);
}

/******************************************************************************
 *
 * slld_PPOp - Performs a Page Programming Operation.
 *
 * This function programs location in a page to the specified data.
 * Function issues all required commands and polls for completion.
 * Data size to program must be within PAZE_SIZE.
 *
 * RETURNS: SLLD_OK, SLLD_E_DEVICE_SOFTWARE_PROTECTED or SLLD_E_HAL_ERROR
 *
 */
SLLD_STATUS slld_PPOp
(
        BYTE  device_num,            /* device number */
        ADDRESS     sys_addr,                 /* device address given by system */
        BYTE       *data_buf,                 /* variable containing data to program */
        BYTECOUNT   len_in_bytes,             /* number of bytes to program */
        DEVSTATUS  *dev_status_ptr            /* variable to store device status */
)
{
    SLLD_STATUS status = SLLD_OK;

    if(len_in_bytes > PAGE_SIZE)
    {
        status = SLLD_ERROR;  /* Data Bytes are larger than Page_Size */
        return(status);
    }

    status = slld_WRENCmd(device_num);
    if(status != SLLD_OK)
        return(status);

    status = slld_PPCmd(device_num, sys_addr, data_buf, len_in_bytes);
    if(status != SLLD_OK)
        return(status);

    status = slld_Poll(device_num, dev_status_ptr);
    if(status != SLLD_OK)
        return(status);

    status = slld_WRDICmd(device_num); /* just in case PPOp is operated on protected area */

    return(status);
}










/******************************************************************************
 *
 * slld_Poll - Polls flash device for embedded operation completion
 *
 * This function polls the Flash device to determine when an embedded
 * operation is finished.
 *
 * RETURNS: SLLD_OK, SLLD_E_DEVICE_SOFTWARE_PROTECTED or SLLD_E_HAL_ERROR
 *
 */
SLLD_STATUS slld_Poll
(
        BYTE  device_num,            /* device number */
        DEVSTATUS  *dev_status_ptr              /* variable to store device status */
)
{
    SLLD_STATUS       status = SLLD_OK;

    *dev_status_ptr = dev_status_unknown;

#ifdef INCLUDE_SOFTWARE_PROTECT_STATUS_CHECK
    /* check if target device is software protected */
    if(sys_software_protect_status != FLASH_SOFTWARE_UNPROTECTED)
    {
        status = SLLD_E_DEVICE_SOFTWARE_PROTECTED;
        return(status);
    }
#endif

    do
    {
        status = slld_StatusGet(device_num, dev_status_ptr);
        if(status != SLLD_OK)
            return(status);
    }
    while(*dev_status_ptr == dev_busy);

    return(status);
}

/******************************************************************************
 *
 * slld_StatusGet - Determines Flash Status
 *
 * This function gets the device status from the SPI flash.
 *
 * RETURNS: SLLD_OK, SLLD_E_DEVICE_SOFTWARE_PROTECTED or SLLD_E_HAL_ERROR
 *
 */
SLLD_STATUS slld_StatusGet
(
        BYTE  device_num,            /* device number */
        DEVSTATUS*  dev_status_ptr                  /* variable to store device status */
)
{
    SLLD_STATUS  status = SLLD_OK;
    BYTE poll_data;

#ifdef INCLUDE_SOFTWARE_PROTECT_STATUS_CHECK
    /* check if target device is software protected */
    if(sys_software_protect_status != FLASH_SOFTWARE_UNPROTECTED)
    {
        status = SLLD_E_DEVICE_SOFTWARE_PROTECTED;
        return(status);
    }
#endif

    status = slld_RDSRCmd(device_num, &poll_data);
    if(status != SLLD_OK)
        return(status);

    // Check if steady state
    if (poll_data & B0_MASK) {                     // If b0 = 1 then device is busy
        // If b5 = 1 then there was an erase error
        if ((STATUS_ERROR_FLAGS & HAS_STATBIT5_ERROR) && (poll_data & B5_MASK))
            *dev_status_ptr = dev_erase_error;
        // If b6 = 1 then there was a program error
        else if ((STATUS_ERROR_FLAGS & HAS_STATBIT6_ERROR) && (poll_data & B6_MASK))
            *dev_status_ptr = dev_program_error;
        else
            *dev_status_ptr = dev_busy;
    }
    else
    {
        // If b0 = 0 then device is not busy
        *dev_status_ptr = dev_not_busy;
    }

    return (status);
}

#ifdef INCLUDE_SOFTWARE_PROTECT_STATUS_CHECK
/******************************************************************************
 *
 * slld_SoftwareProtectStatusGet - Determines Flash Software Protection Status
 *
 * This function determines the flash software protection status.
 *
 * RETURNS: SLLD_OK.
 *
 */
SLLD_STATUS slld_SoftwareProtectStatusGet    ///xbill ????, how to add device_num?
(
        DEV_SOFTWARE_PROTECT_STATUS*  dev_softwareprotect_status_ptr  /* variable to store device software protect status */
)
{
    SLLD_STATUS status = SLLD_OK;

    *dev_softwareprotect_status_ptr = sys_software_protect_status;

    return(status);
}
#endif /* INCLUDE_SOFTWARE_PROTECT_STATUS_CHECK */

































SLLD_STATUS slld_ReadOp
(
        BYTE  device_num,            /* device number */
        ADDRESS   sys_addr,          /* device address given by system */
        BYTE     *target,            /* variable in which to store read data */
        BYTECOUNT len_in_bytes       /* number of bytes to read */
)
{
    SLLD_STATUS         status = SLLD_OK;
    unsigned long        length;
#ifdef INCLUDE_SOFTWARE_PROTECT_STATUS_CHECK
    /* check if target device is software protected */
    if(sys_software_protect_status != FLASH_SOFTWARE_UNPROTECTED)
    {
        status = SLLD_E_DEVICE_SOFTWARE_PROTECTED;
        return(status);
    }
#endif

    while (len_in_bytes)
    {
#ifdef CONTROLLER_BUFFER_SIZE
        if (len_in_bytes < CONTROLLER_BUFFER_SIZE)
            length = len_in_bytes;
        else
            length = CONTROLLER_BUFFER_SIZE;
#else
        length = len_in_bytes;
#endif

#ifdef USE_4B_ADDR_CMDS
#ifdef USE_FAST_READ
        status = FLASH_RD(device_num,SPI_FAST_READ_4B_CMD, sys_addr, target, length);
#elif defined USE_DUAL_READ
        status = FLASH_RD(device_num,SPI_DUALIO_RD_4B_CMD, sys_addr, target, length);
#elif defined USE_QUAD_READ
        status = FLASH_RD(device_num,SPI_QUADIO_RD_4B_CMD, sys_addr, target, length);
#else
        status = FLASH_RD(device_num,SPI_READ_4B_CMD, sys_addr, target, length);
#endif
#else //#ifdef USE_4B_ADDR_CMDS
#ifdef USE_FAST_READ
        status = FLASH_RD(device_num,SPI_FAST_READ_CMD, sys_addr, target, length);
#elif defined USE_DUAL_READ
        status = FLASH_RD(device_num,SPI_DUALIO_RD_CMD, sys_addr, target, length);
#elif defined USE_QUAD_READ
        status = FLASH_RD(device_num,SPI_QUADIO_RD_CMD, sys_addr, target, length);
#else
        status = FLASH_RD(device_num,SPI_READ_CMD, sys_addr, target, length);
#endif
#endif

        len_in_bytes -= length;
        sys_addr += length;
        target += length;
    }

    return(status);
}


SLLD_STATUS slld_WriteOp
(
        BYTE  device_num,            /* device number */
        ADDRESS     sys_addr,                 /* device address given by system */
        BYTE       *data_buf,                 /* variable containing data to program */
        BYTECOUNT   len_in_bytes,             /* number of bytes to program */
        DEVSTATUS  *dev_status_ptr            /* variable to store device status */
)
{
    SLLD_STATUS status = SLLD_OK;
    unsigned long length;

    while(len_in_bytes)
    {

        if(len_in_bytes < PAGE_SIZE)
            length = len_in_bytes;
        else
            length = PAGE_SIZE;

        status = slld_WRENCmd(device_num);
        if(status != SLLD_OK)
            return(status);
#ifdef USE_4B_ADDR_CMDS
#ifdef USE_QUAD_WRITE
        status = slld_QPP_4BCmd(device_num, sys_addr, data_buf, length);
#else
        status = slld_PP_4BCmd(device_num, sys_addr, data_buf, length);
#endif
#else //#ifdef USE_4B_ADDR_CMDS
#ifdef USE_QUAD_WRITE
        status = slld_QPPCmd(device_num, sys_addr, data_buf, length);
#else
        status = slld_PPCmd(device_num, sys_addr, data_buf, length);
#endif
#endif

        if(status != SLLD_OK)
            return(status);

        status = slld_Poll(device_num, dev_status_ptr);
        if(status != SLLD_OK)
            return(status);

        status = slld_WRDICmd(device_num);
        if(status != SLLD_OK)
            return(status);

        len_in_bytes -= length;
        sys_addr += length;
        data_buf += length;
    }

    return(status);
}






/******************************************************************************
 *
 * slld_GetDevNumFromAddr -help user to get device number from address when target device with multiple chip select
 *
 * sys_addr       device address given by system, sys_addr may includes system base address + device offset address
 * device_num   device number based on sys_addr
 *     device_num = 0x00, drive CS[0]
 *     device_num = 0x01, drive CS[1]
 * RETURNS: SLLD_OK, SLLD_E_DEVICE_SOFTWARE_PROTECTED or SLLD_E_HAL_ERROR
 *
 */
SLLD_STATUS slld_GetDevNumFromAddr
(
        ADDRESS     sys_addr,                  /* device address given by system */
        BYTE  *device_num                        /* device number */
)
{
    SLLD_STATUS status = SLLD_OK;

    *device_num = (sys_addr & BASE_ADDR_MASK) >> BASE_ADDR_SHIFT_BIT;
    
    return status;
}
































